@@ -6,6 +6,8 @@
 
 from __future__ import print_function
 
+import glob
+import json
 import os
 import shutil
 import sys
@@ -25,17 +27,19 @@ from traitlets.utils.importstring import import_item
 
 from tornado.log import LogFormatter
 
-from . import (
-    get_labextension_manifest_data_by_folder, semver
-)
+from .semver import satisfies
+
 
 # Constants for pretty print extension listing function.
 # Window doesn't support coloring in the commandline
-GREEN_ENABLED = '\033[32m enabled \033[0m' if os.name != 'nt' else 'enabled '
+GREEN_OK = '\033[32mOK\033[0m' if os.name != 'nt' else 'ok'
+RED_X = '\033[31mX\033[0m' if os.name != 'nt' else 'X'
+
+GREEN_ENABLED = '\033[32menabled \033[0m' if os.name != 'nt' else 'enabled '
 RED_DISABLED = '\033[31mdisabled\033[0m' if os.name != 'nt' else 'disabled'
 
-GREEN_OK = '\033[32mOK\033[0m' if os.name != 'nt' else 'ok'
-RED_X = '\033[31m X\033[0m' if os.name != 'nt' else ' X'
+
+CONFIG_NAME = 'jupyter_notebook_config'
 
 #------------------------------------------------------------------------------
 # Public API
@@ -256,7 +260,8 @@ def uninstall_labextension_python(module,
 
 
 def _set_labextension_state(name, state,
-                           user=True, sys_prefix=False, logger=None):
+                           user=True, sys_prefix=False, logger=None,
+                           python_module=None):
     """Set whether the JupyterLab frontend should use the named labextension
 
     Returns True if the final state is the one requested.
@@ -272,24 +277,28 @@ def _set_labextension_state(name, state,
         Whether to update the sys.prefix, i.e. environment. Will override
         `user`.
     logger : Jupyter logger [optional]
-        Logger instance to use
+        Logger instance to use [optional]
+    python_module: string
+        The name of the python module associated with the extension.
     """
     user = False if sys_prefix else user
-    config_dir = os.path.join(
-        _get_config_dir(user=user, sys_prefix=sys_prefix), 'labconfig')
-    cm = BaseJSONConfigManager(config_dir=config_dir)
+    cfg = _read_config_data(user=user, sys_prefix=sys_prefix)
     if logger:
         logger.info("{} extension {}...".format(
             "Enabling" if state else "Disabling",
             name
         ))
-    cfg = cm.get("jupyterlab_config")
     labextensions = (
         cfg.setdefault("LabApp", {})
         .setdefault("labextensions", {})
     )
 
-    old_enabled = labextensions.get(name, None)
+    old_state = labextensions.get(name, None)
+    if old_state is None:
+        old_state = dict(enabled=False)
+    elif isinstance(old_state, bool):
+        old_state = dict(enabled=old_state)
+    old_enabled = old_state['enabled']
     new_enabled = state if state is not None else not old_enabled
 
     if logger:
@@ -298,15 +307,16 @@ def _set_labextension_state(name, state,
         else:
             logger.info(u"Disabling: %s" % (name))
 
-    labextensions[name] = new_enabled
+    labextensions[name] = dict(
+        enabled=new_enabled,
+        python_module=python_module
+    )
 
-    if logger:
-        logger.info(u"- Writing config: {}".format(config_dir))
-
-    cm.update("jupyterlab_config", cfg)
+    _write_config_data(cfg, user=user, sys_prefix=sys_prefix, logger=logger)
 
     if new_enabled:
-        validate_labextension(name, logger=logger)
+        full_dest = find_labextension(name)
+        validate_labextension_folder(name, full_dest, logger=logger)
 
     return old_enabled == state
 
@@ -337,7 +347,8 @@ def _set_labextension_state_python(state, module, user, sys_prefix,
     return [_set_labextension_state(name=labext["name"],
                                    state=state,
                                    user=user, sys_prefix=sys_prefix,
-                                   logger=logger)
+                                   logger=logger,
+                                   python_module=module)
             for labext in labexts]
 
 
@@ -436,24 +447,24 @@ def disable_labextension_python(module, user=True, sys_prefix=False,
                                          logger=logger)
 
 
-def validate_labextension(name, logger=None):
-    """Validate a named labextension.
+def find_labextension(name):
+    """Find a labextension path
 
     Looks across all of the labextension directories.
 
-    Returns a list of warnings.
+    Returns the first path where the extension is found,
+    or None if not found.
 
     Parameters
     ----------
     name : str
         The name of the extension.
-    logger : Jupyter logger [optional]
-        Logger instance to use
     """
     for exts in _labextension_dirs():
         full_dest = os.path.join(exts, name)
         if os.path.exists(full_dest):
-            return validate_labextension_folder(name, full_dest, logger)
+            return full_dest
+    return None
 
 
 def validate_labextension_folder(name, full_dest, logger=None):
@@ -469,29 +480,27 @@ def validate_labextension_folder(name, full_dest, logger=None):
     logger : Jupyter logger [optional]
         Logger instance to use
     """
-    if logger:
-        logger.info("    - Validating...")
-
     infos = []
     warnings = []
 
-    hasFiles = True
-    hasEntry = False
-    versonCompatible = True
+    has_files = []
+    has_entry = False
+    version_compatible = []
 
     data = get_labextension_manifest_data_by_folder(full_dest)
     for manifest in data.values():
         if ('entry' in manifest and 'modules' in manifest):
             if (manifest['entry'] in manifest['modules']):
-                hasEntry = True
+                has_entry = True
         files = manifest.get('files', [])
         if not files:
-            hasFiles = False
-        for fname in files:
-            path = os.path.join(full_dest, fname)
-            if not os.path.exists(path):
-                hasFiles = False
-        # add check here for version compat
+            has_files.append("No 'files' key in manifest")
+        else:
+            for fname in files:
+                path = os.path.join(full_dest, fname)
+                if not os.path.exists(path):
+                    has_files.append("File in manifest does not exist: {}".format(path))
+
         for (mod, deps) in manifest.get('modules', {}).items():
             # ignore jupyterlab modules since
             # they are only semver compatibile with their own version
@@ -500,39 +509,77 @@ def validate_labextension_folder(name, full_dest, logger=None):
             for dep in deps:
                 if dep.startswith('jupyterlab@'):
                     dep = dep.split('/')[0].split('@')[-1]
-                    if not semver.satisfies(__version__, dep, False):
-                        versonCompatible = False
+                    if not satisfies(__version__, dep, False):
+                        version_compatible.append('Expects JupyterLab version %s from packaged module %s'%(dep, mod))
+                        break
 
-    entry_msg = u"     {} has entry point?"
+    indent = "        "
+    subindent = indent + "    "
+    entry_msg = indent + u"{} has entry point in manifest"
     name = os.path.basename(full_dest)
-    if hasEntry:
+    if has_entry:
         infos.append(entry_msg.format(GREEN_OK))
     else:
         warnings.append(entry_msg.format(RED_X))
 
-    files_msg = u"     {} has necessary files?"
-    if hasFiles:
+    files_msg = indent+u"{} has necessary files"
+    if len(has_files)==0:
         infos.append(files_msg.format(GREEN_OK))
     else:
         warnings.append(files_msg.format(RED_X))
+        for m in has_files:
+            warnings.append(subindent+m)
 
-    version_msg = u"     {} is version compatible?"
-    if versonCompatible:
-        infos.append(version_msg.format(GREEN_ENABLED))
+    if len(version_compatible)==0:
+        infos.append(indent+"{} is version compatible with installed JupyterLab version {}".format(GREEN_OK, __version__))
     else:
-        warnings.append(version_msg.format(RED_X))
+        warnings.append(indent+"{} is not version compatible with installed JupyterLab version {}".format(RED_X, __version__))
+        for m in version_compatible:
+            warnings.append(subindent+m)
 
-    post_mortem = u"      {} {} {}"
-    if logger:
-        if warnings:
-            [logger.info(info) for info in infos]
+    if logger and warnings:
+            #[logger.info(info) for info in infos]
             [logger.warn(warning) for warning in warnings]
-        else:
-            logger.info(post_mortem.format(name, "", GREEN_OK))
 
     return warnings
 
 
+def get_labextension_manifest_data_by_folder(folder):
+    """Get the manifest data for a given lab extension folder
+    """
+    manifest_files = glob.glob(os.path.join(folder, '*.manifest'))
+    manifests = {}
+    for file in manifest_files:
+        with open(file) as fid:
+            manifest = json.load(fid)
+        manifests[manifest['name']] = manifest
+    return manifests
+
+
+def get_labextension_manifest_data_by_name(name):
+    """Get the manifest data for a given lab extension folder
+    """
+    for exts in _labextension_dirs():
+        full_dest = os.path.join(exts, name)
+        if os.path.exists(full_dest):
+            return get_labextension_manifest_data_by_folder(full_dest)
+
+
+def get_labextension_config_python(module):
+    """Get the labextension configuration data associated  with a Python module. 
+
+    Parameters
+    -----------
+    module : str
+    Importable Python module exposing the
+    magic-named `_jupyter_labextension_config` function
+    """
+    m = import_item(module)
+    if not hasattr(m, '_jupyter_labextension_config'):
+        return {}
+    return m._jupyter_labextension_config()
+
+
 #----------------------------------------------------------------------
 # Applications
 #----------------------------------------------------------------------
@@ -844,24 +891,29 @@ class ListLabExtensionsApp(BaseLabExtensionApp):
     
     def list_labextensions(self):
         """List all the labextensions"""
-        config_dirs = [os.path.join(p, 'labconfig') for p in jupyter_config_path()]
-        
         print("Known labextensions:")
-        
-        for config_dir in config_dirs:
+
+        for config_dir in jupyter_config_path():
             cm = BaseJSONConfigManager(parent=self, config_dir=config_dir)
-            data = cm.get("jupyterlab_config")
+            data = cm.get(CONFIG_NAME)
             labextensions = (
                 data.setdefault("LabApp", {})
                 .setdefault("labextensions", {})
             )
             if labextensions:
                 print(u'config dir: {}'.format(config_dir))
-            for name, enabled in labextensions.items():
-                print(u'    {} {}'.format(
+            for name, config in sorted(labextensions.items()):
+                if isinstance(config, bool):
+                    config = dict(enabled=config)
+                enabled = config['enabled']
+                full_dest = find_labextension(name)
+                print(u'    {} {}: {}'.format(
                               name,
-                              GREEN_ENABLED if enabled else RED_DISABLED))
-                validate_labextension(name, self.log)
+                              GREEN_ENABLED if enabled else RED_DISABLED,
+                              full_dest if not None else RED_X+" Files not found"
+                              ))
+                if full_dest is not None:
+                    validate_labextension_folder(name, full_dest, self.log)
 
     def start(self):
         """Perform the App's functions as configured"""
@@ -1075,10 +1127,10 @@ def _read_config_data(user=False, sys_prefix=False):
     """
     config_dir = _get_config_dir(user=user, sys_prefix=sys_prefix)
     config_man = BaseJSONConfigManager(config_dir=config_dir)
-    return config_man.get('jupyterlab_config')
+    return config_man.get(CONFIG_NAME)
 
 
-def _write_config_data(data, user=False, sys_prefix=False):
+def _write_config_data(data, user=False, sys_prefix=False, logger=None):
     """Update the config for the current context
 
     Parameters
@@ -1089,10 +1141,14 @@ def _write_config_data(data, user=False, sys_prefix=False):
         Get the user's .jupyter config directory
     sys_prefix : bool [default: False]
         Get sys.prefix, i.e. ~/.envs/my-env/etc/jupyter
+    logger: logger instance
+        The logger instance.
     """
     config_dir = _get_config_dir(user=user, sys_prefix=sys_prefix)
+    if logger:
+        logger.info(u"- Writing config: {}".format(config_dir))
     config_man = BaseJSONConfigManager(config_dir=config_dir)
-    config_man.update('jupyterlab_config', data)
+    config_man.update(CONFIG_NAME, data)
 
 
 if __name__ == '__main__':
